// Code generated by generate.go. DO NOT EDIT.

package stylecheck

import (
	"honnef.co/go/tools/analysis/lint"
	"honnef.co/go/tools/stylecheck/st1000"
	"honnef.co/go/tools/stylecheck/st1001"
	"honnef.co/go/tools/stylecheck/st1003"
	"honnef.co/go/tools/stylecheck/st1005"
	"honnef.co/go/tools/stylecheck/st1006"
	"honnef.co/go/tools/stylecheck/st1008"
	"honnef.co/go/tools/stylecheck/st1011"
	"honnef.co/go/tools/stylecheck/st1012"
	"honnef.co/go/tools/stylecheck/st1013"
	"honnef.co/go/tools/stylecheck/st1015"
	"honnef.co/go/tools/stylecheck/st1016"
	"honnef.co/go/tools/stylecheck/st1017"
	"honnef.co/go/tools/stylecheck/st1018"
	"honnef.co/go/tools/stylecheck/st1019"
	"honnef.co/go/tools/stylecheck/st1020"
	"honnef.co/go/tools/stylecheck/st1021"
	"honnef.co/go/tools/stylecheck/st1022"
	"honnef.co/go/tools/stylecheck/st1023"
)

var Analyzers = []*lint.Analyzer{
	st1000.SCAnalyzer,
	st1001.SCAnalyzer,
	st1003.SCAnalyzer,
	st1005.SCAnalyzer,
	st1006.SCAnalyzer,
	st1008.SCAnalyzer,
	st1011.SCAnalyzer,
	st1012.SCAnalyzer,
	st1013.SCAnalyzer,
	st1015.SCAnalyzer,
	st1016.SCAnalyzer,
	st1017.SCAnalyzer,
	st1018.SCAnalyzer,
	st1019.SCAnalyzer,
	st1020.SCAnalyzer,
	st1021.SCAnalyzer,
	st1022.SCAnalyzer,
	st1023.SCAnalyzer,
}
